# Copyright 2017, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import aioredis
import wrapt

from qubit.opencensus.trace import asyncio_context

log = logging.getLogger(__name__)

MODULE_NAME = 'aioredis'

CONNECTION_WRAP_METHODS = 'execute'
CONNECTION_CLASS_NAME = 'RedisConnection'


def trace_integration(tracer=None):
    # Wrap Session class
    wrapt.wrap_function_wrapper(
        MODULE_NAME, 'RedisConnection.execute', wrap_execute)


async def wrap_execute(wrapped, instance, args, kwargs):
    """Wrap the session function to trace it."""
    command = args[0]
    _tracer = asyncio_context.get_opencensus_tracer()
    _span = _tracer.start_span()
    _span.name = '[aioredis]{}'.format(command)

    # Add the requests url to attributes
    try:
        result = await wrapped(*args, **kwargs)
        _span.finish()
    except Exception as e: 
        _span.add_attribute('error', True)
        _span.add_attribute('error.message', str(e))
        _span.finish()
        raise e

    return result



# Copyright 2017, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import aiohttp

from qubit.opencensus.trace import asyncio_context

log = logging.getLogger(__name__)

MODULE_NAME = 'aiohttp'


def trace_integration(tracer=None, propagator=None):
    """Wrap the requests library to trace it."""
    log.info('Integrated module: {}'.format(MODULE_NAME))
    # Wrap the aiohttp functions
    aiohttp_func = getattr(aiohttp.ClientSession, '_request')
    wrapped = wrap_aiohttp(aiohttp_func, propagator=propagator)
    setattr(aiohttp.ClientSession, aiohttp_func.__name__, wrapped)


def wrap_aiohttp(aiohttp_func, propagator=None):
    """Wrap the aiohttp function to trace it."""
    async def call(*args, **kwargs):
            _tracer = asyncio_context.get_opencensus_tracer()
            _span = _tracer.start_span()
            _span.name = '[aiohttp]{}'.format(args[1])
            _span.add_attribute('aiohttp/url', str(args[2]))

            if propagator is not None:
                span_context = _span.context_tracer.span_context
                headers = propagator.to_headers(span_context)
                if 'headers' not in kwargs:
                    kwargs['headers'] = {}
                for k, v in headers.items():
                    kwargs['headers'][k] = v

            try:
                response = await aiohttp_func(*args, **kwargs)

                _span.add_attribute(
                    'aiohttp/status_code', str(response.status))
                _span.add_attribute(
                    'aiohttp/status_reason', str(response.reason))

                if response.status >= 500:
                    _span.add_attribute(
                        'aiohttp/error', True)

                _span.finish()

                return response
            except Exception as e:
                _span.add_attribute(
                    'aiohttp/error', True)
                _span.add_attribute(
                    'aiohttp/error,message', str(e))
                _span.finish()
                raise e

    return call



# Copyright 2017, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import aiotask_context as context

from opencensus.trace.tracers import noop_tracer

_TRACER_KEY = 'opencensus.io/trace'
_ATTRS_KEY = 'opencensus.io/attrs'
_CURRENT_SPAN_KEY = 'opencensus.io/current-span'


def get_opencensus_tracer():
    return context.get(_TRACER_KEY, default=noop_tracer.NoopTracer())


def set_opencensus_tracer(tracer):
    """Add the tracer to thread local."""
    context.set(key=_TRACER_KEY, value=tracer)


def set_opencensus_attr(attr_key, attr_value):
    attrs = context.get(_ATTRS_KEY, {})
    attrs[_ATTRS_KEY] = attr_value
    context.set(key=_ATTRS_KEY, value=attrs)


def get_opencensus_attr(attr_key):
    attrs = context.get(_ATTRS_KEY, None)
    if attrs is not None:
        return attrs.get(_ATTRS_KEY)
    return None


def get_current_span():
    return context.get(_CURRENT_SPAN_KEY, None)


def set_current_span(current_span):
    context.set(key=_CURRENT_SPAN_KEY, value=current_span)


def clear():
    """Clear the thread local, used in test."""
    context.clear()



# Copyright 2017, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A setup module for Open Source Census Instrumentation Library"""

import io
from setuptools import setup, find_packages

extras = {}

install_requires = []

packages = [
    package for package in find_packages()
    if package.startswith('qubit')]

namespaces = ['qubit']

setup(
    name='qubit-opencensus',
    version='0.1.0',
    author='Qubit',
    author_email='tristan@qubit.com',
    classifiers=[
        'Intended Audience :: Developers',
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    description='Various extensions for OpenCensus',
    include_package_data=True,
    long_description=open('README.rst').read(),
    install_requires=install_requires,
    extras_require=extras,
    license='Apache-2.0',
    packages=packages,
    namespace_packages=namespaces,
    url='https://github.com/QubitProducts/opencensus-utils')



# Copyright 2017, OpenCensus Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
import logging
import sys

from sanic.exceptions import SanicException

from google.rpc import code_pb2

from opencensus.trace import attributes_helper
from opencensus.trace import stack_trace
from opencensus.trace import status
from opencensus.trace.exporters import print_exporter
from opencensus.trace.exporters.transports import sync
from opencensus.trace.ext import utils
from opencensus.trace.propagation import google_cloud_format
from opencensus.trace.samplers import always_on, probability
from qubit.opencensus.trace import asyncio_context
from qubit.opencensus.trace.tracers import asyncio_context_tracer as tracer_module


_SANIC_TRACE_HEADER = 'X_CLOUD_TRACE_CONTEXT'

HTTP_METHOD = attributes_helper.COMMON_ATTRIBUTES['HTTP_METHOD']
HTTP_URL = attributes_helper.COMMON_ATTRIBUTES['HTTP_URL']
HTTP_STATUS_CODE = attributes_helper.COMMON_ATTRIBUTES['HTTP_STATUS_CODE']

BLACKLIST_PATHS = 'BLACKLIST_PATHS'
GCP_EXPORTER_PROJECT = 'GCP_EXPORTER_PROJECT'
SAMPLING_RATE = 'SAMPLING_RATE'
TRANSPORT = 'TRANSPORT'
ZIPKIN_EXPORTER_SERVICE_NAME = 'ZIPKIN_EXPORTER_SERVICE_NAME'
ZIPKIN_EXPORTER_HOST_NAME = 'ZIPKIN_EXPORTER_HOST_NAME'
ZIPKIN_EXPORTER_PORT = 'ZIPKIN_EXPORTER_PORT'

log = logging.getLogger(__name__)


class SanicMiddleware(object):
    DEFAULT_SAMPLER = always_on.AlwaysOnSampler
    DEFAULT_EXPORTER = print_exporter.PrintExporter
    DEFAULT_PROPAGATOR = google_cloud_format.GoogleCloudFormatPropagator

    """sanic middleware to automatically trace requests.

    :type app: :class: `~sanic.sanic`
    :param app: A sanic application.

    :type blacklist_paths: list
    :param blacklist_paths: Paths that do not trace.

    :type sampler: :class: `type`
    :param sampler: Class for creating new Sampler objects. It should extend
                    from the base :class:`.Sampler` type and implement
                    :meth:`.Sampler.should_sample`. Defaults to
                    :class:`.AlwaysOnSampler`. The rest options are
                    :class:`.AlwaysOffSampler`, :class:`.FixedRateSampler`.

    :type exporter: :class: `type`
    :param exporter: Class for creating new exporter objects. Default to
                     :class:`.PrintExporter`. The rest option is
                     :class:`.FileExporter`.

    :type propagator: :class: 'type'
    :param propagator: Class for creating new propagator objects. Default to
                       :class:`.GoogleCloudFormatPropagator`. The rest option
                       are :class:`.BinaryFormatPropagator`,
                       :class:`.TextFormatPropagator` and
                       :class:`.TraceContextPropagator`.
    """
    def __init__(self, app=None, blacklist_paths=None, sampler=None,
                 exporter=None, propagator=None):
        self.app = app
        self.blacklist_paths = blacklist_paths
        self.sampler = sampler
        self.exporter = exporter
        self.propagator = propagator

        if self.app is not None:
            self.init_app(app)

    def init_app(self, app):
        self.app = app

        # get settings from app config
        settings = self.app.config.get('OPENCENSUS_TRACE', {})

        self.sampler = (self.sampler
                        or settings.get('SAMPLER',
                                        self.DEFAULT_SAMPLER))
        self.exporter = (self.exporter
                         or settings.get('EXPORTER',
                                         self.DEFAULT_EXPORTER))
        self.propagator = (self.propagator
                           or settings.get('PROPAGATOR',
                                           self.DEFAULT_PROPAGATOR))

        # get params from app config
        params = self.app.config.get('OPENCENSUS_TRACE_PARAMS', {})

        self.blacklist_paths = params.get(BLACKLIST_PATHS,
                                          self.blacklist_paths)

        # Initialize the sampler
        if not inspect.isclass(self.sampler):
            pass  # handling of instantiated sampler
        elif self.sampler.__name__ == 'ProbabilitySampler':
            _rate = params.get(SAMPLING_RATE,
                               probability.DEFAULT_SAMPLING_RATE)
            self.sampler = self.sampler(_rate)
        else:
            self.sampler = self.sampler()

        transport = params.get(TRANSPORT, sync.SyncTransport)

        # Initialize the exporter
        if not inspect.isclass(self.exporter):
            pass  # handling of instantiated exporter
        elif self.exporter.__name__ == 'StackdriverExporter':
            _project_id = params.get(GCP_EXPORTER_PROJECT, None)
            self.exporter = self.exporter(
                project_id=_project_id,
                transport=transport)
        elif self.exporter.__name__ == 'ZipkinExporter':
            _zipkin_service_name = params.get(
                ZIPKIN_EXPORTER_SERVICE_NAME, 'my_service')
            _zipkin_host_name = params.get(
                ZIPKIN_EXPORTER_HOST_NAME, 'localhost')
            _zipkin_port = params.get(
                ZIPKIN_EXPORTER_PORT, 9411)
            self.exporter = self.exporter(
                service_name=_zipkin_service_name,
                host_name=_zipkin_host_name,
                port=_zipkin_port,
                transport=transport)
        else:
            self.exporter = self.exporter(transport=transport)

        # Initialize the propagator
        if inspect.isclass(self.propagator):
            self.propagator = self.propagator()

        @app.middleware('request')
        async def trace_request(request):
            self.do_trace_request(request)

        @app.middleware('response')
        async def trace_response(request, response):
            self.do_trace_response(request, response)

        @app.exception(SanicException)
        async def trace_exception(request, exception):
            self.do_trace_exception(request, exception)

    def do_trace_request(self, request):
        if utils.disable_tracing_url(request.url, self.blacklist_paths):
            return

        try:
            span_context = self.propagator.from_headers(request.headers)
            tracer = tracer_module.ContextTracer(
                span_context=span_context,
                exporter=self.exporter)

            span = tracer.start_span()

            # Set the span name as the name of the current module name
            span.name = '[{}]{}'.format(
                request.method,
                request.url)
            span.add_attribute(
                HTTP_METHOD, request.method)
            span.add_attribute(HTTP_URL, request.url)
            request['tracer'] = tracer
            request['span'] = span
            asyncio_context.set_opencensus_tracer(tracer)
        except Exception:  # pragma: NO COVER
            log.error('Failed to trace request', exc_info=True)

    def do_trace_response(self, request, response):
        """A function to be run after each request.
        """
        # Do not trace if the url is blacklisted
        if utils.disable_tracing_url(request.url, self.blacklist_paths):
            return response

        try:
            tracer = request['tracer']
            span = request['span']
            span.add_attribute(
                HTTP_STATUS_CODE,
                str(response.status))
            if response.status >= 500:
                span.add_attribute('error', True)

            span.finish()
            tracer.finish()
        except Exception:  # pragma: NO COVER
            log.error('Failed to trace request', exc_info=True)

    def do_trace_exception(self, request, exception):
        # Do not trace if the url is blacklisted
        if utils.disable_tracing_url(request.url, self.blacklist_paths):
            return

        try:
            tracer = request['tracer']
            span = request['span']
            if exception is not None:
                span.add_attribute('error', True)
                span.add_attribute('error.message', str(exception))

            span.finish()
            tracer.finish()
        except Exception:  # pragma: NO COVER
            log.error('Failed to trace request', exc_info=True)
